# generated by datamodel-codegen:
#   filename:  comment_schema.json
#   timestamp: 2024-02-07T08:09:26+00:00

from enum import Enum
from typing import Optional

from pydantic import BaseModel, Extra, Field

from omnipy.data.model import Model


class FieldType(Enum):
    Comment = 'Comment'


class IsaCommentSchema(BaseModel):
    class Config:
        extra = Extra.forbid
        use_enum_values = True

    field_id: Optional[str] = Field(None, alias='@id')
    field_context: Optional[str] = Field(None, alias='@context')
    field_type: Optional[FieldType] = Field(None, alias='@type')
    name: Optional[str] = None
    value: Optional[str] = None


class IsaCommentModel(Model[IsaCommentSchema]):
    ...
